/*
 * This is free software
 *
 * getsadcvalue.c - part of xsar program 
 *
 * Author:   Petr Bravenec
 *
 * Version:  0.2  17-Jul-1996
 *
 */

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include "xsar.h"
#include "kmem_seek.h"

Sar  sar[SARNUM];
int  sarend;
int  sarnum;

#ifdef SVR4 
sa_t sa_old;
sa_t sa_new;

/**************************************************************************/
void getsadcvalue (void) {
    memcpy (&sa_old,&sa_new,sizeof(sa_t));
    lseek  (kmem_fd,seek_lbolt,SEEK_SET);
    read   (kmem_fd,&sa_new.lbolt,sizeof(sa_old.lbolt));
    lseek  (kmem_fd,seek_sysinfo,SEEK_SET);
    read   (kmem_fd,&sa_new.si,sizeof(sa_old.si));
    lseek  (kmem_fd,seek_minfo,SEEK_SET);
    read   (kmem_fd,&sa_new.mi,sizeof(sa_old.mi));
    lseek  (kmem_fd,seek_vminfo,SEEK_SET);
    read   (kmem_fd,&sa_new.vmi,sizeof(sa_old.vmi));
    lseek  (kmem_fd,seek_freemem,SEEK_SET);
    read   (kmem_fd,&sa_new.freemem,sizeof(sa_old.freemem));
}

/**************************************************************************/
void readvalues (void) {
    Sar msar;
    int ticks = sa_new.lbolt - sa_old.lbolt;
    getsadcvalue();
    msar.runque   = 10000*(sa_new.si.runque-sa_old.si.runque)/ticks;
    msar.freeswap = 100 * sa_new.mi.freeswap/512;
    msar.free     = 100 * sa_new.freemem/128;
    msar.freemem  = msar.free - msar.freeswap;
    msar.cpu_idle = (100*(sa_new.si.cpu[CPU_IDLE]   - sa_old.si.cpu[CPU_IDLE])  )/delay;
    msar.cpu_user = (100*(sa_new.si.cpu[CPU_USER]   - sa_old.si.cpu[CPU_USER])  )/delay;
    msar.cpu_sys  = (100*(sa_new.si.cpu[CPU_KERNEL] - sa_old.si.cpu[CPU_KERNEL]))/delay;
    msar.cpu_wait = (100*(sa_new.si.cpu[CPU_WAIT]   - sa_old.si.cpu[CPU_WAIT])  )/delay;
    msar.pg_in    = sa_new.si.swapin  - sa_old.si.swapin;
    msar.pg_out   = sa_new.si.swapout - sa_old.si.swapout;
#endif /* SVR4 */
/**************************************************************************/
#ifdef linux
void readvalues (void) {
    Sar   msar;
    FILE *fd;
    char  buf[256];
    {
    double a,b,r;
    fd = fopen ("/proc/loadavg","rt");
    fgets  (buf,256,fd);
    sscanf (buf,"%lf %lf %lf",&r,&a,&b);
    msar.runque = r * 100.;
    fclose (fd);
    } {
    int a,b,c,d,e;
    fd = fopen ("/proc/meminfo","rt");
    while (fgets (buf,256,fd)) {
	if (5==sscanf(buf,"Mem: %i %i %i %i %i",&a,&b,&c,&d,&e)) 
	    msar.freemem = c/10486 + e/10486;
	if (3==sscanf(buf,"Swap: %i %i %i",&a,&b,&c))
	    msar.freeswap = c/10486; 
	}
    msar.free     = msar.freeswap + msar.freemem;
    fclose (fd);
    } {
    int a,b,c,d,e,x;
    static int o_sys, o_user, o_idle, jiffies, o_pgin, o_pgout;
    fd = fopen ("/proc/stat","rt");
    while (fgets (buf, 256, fd)) {
	if (4==sscanf(buf,"cpu %i %i %i %i",&a,&b,&c,&d)) {
            e = a+b+c+d;
            x = e-jiffies;
            msar.cpu_wait = 0; /* I have not found this information in Linux */
            msar.cpu_user = (a - o_user) * 10000 / x ;
            msar.cpu_sys  = (c - o_sys ) * 10000 / x ;
            msar.cpu_idle = (d - o_idle) * 10000 / x ;
            o_sys  = c;
            o_user = a;
            o_idle = d;
            jiffies = e;
            }
	if (2==sscanf(buf,"swap %i %i",&a,&b)) {
            if (sarnum==0) {
                o_pgin=a;
                o_pgout=b;
                }
            msar.pg_in  = a - o_pgin  ;
            msar.pg_out = b - o_pgout ;
            o_pgin = a;
            o_pgout = b;
            }
	}

    fclose (fd);
    }
#endif /* linux */ /* begins common part of function */
    msar.time = time(NULL)/10*10;
    if (sarnum==0) 
	memcpy (&sar[0], &msar, sizeof(Sar)); /* initialise */
    if (++sarend >= SARNUM)
	sarend=0;
    if (++sarnum > SARNUM)
	sarnum=SARNUM;
    memcpy (&sar[sarend], &msar, sizeof(Sar));
/*
    printf ("%4.2f %1.1f+%1.1f=%1.1f %5.1f %5.1f %5.1f %5.1f\n",
       msar.runque, msar.freemem, msar.freeswap, msar.free,
       msar.cpu_idle, msar.cpu_user, msar.cpu_sys, msar.cpu_wait
       );
*/
    }

