/*
 * This is free software
 *
 * info.c - part of xsar program
 * 
 * Author:   Petr Bravenec
 *
 * Version:  0.2  17-Jun-1996
 *
 */
 
#include <stdio.h>
#include <unistd.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <nlist.h>
#include <sys/utsname.h>
#include "xsar.h"
#include "kmem_seek.h"

char   Node[128];
static void handlerInfoWin (Widget, XtPointer, XEvent *, Boolean *);
void   ReadInfo (Arg *);
char  *strlrtrim (char *);

/**************************************************************************/
static void handlerInfoWin (Widget w, XtPointer c, XEvent *event, Boolean *continues) /* ARGSUSED0 */ {
    XtRemoveEventHandler (w, ButtonPressMask, FALSE, handlerInfoWin, NULL);
    XtDestroyWidget ( *((Widget *)c) );
}

/**************************************************************************/
void ShowInfo (Widget w, XtPointer client_data, XtPointer call_data) /* ARGSUSED0 */ {
    Widget    shell, infow;
    Widget    texts, values;
    Arg       attribs[10];
    Position  x,y;
    Dimension height;
    int       n=2;

    XtSetArg    (attribs[0], XtNheight, (XtArgVal)&height);
    XtGetValues (w, attribs, 1); 
    XtTranslateCoords (w, 0, height+2, &x, &y);
    XtSetArg    (attribs[0], XtNx, (XtArgVal) x);
    XtSetArg    (attribs[1], XtNy, (XtArgVal) y);
    ReadInfo    (&attribs[2]);

    shell = XtCreatePopupShell ("InfoWin",transientShellWidgetClass,
		    w, attribs, 2);

    infow = XtCreateManagedWidget ("infow" ,boxWidgetClass,shell,NULL,0);
    texts = XtCreateManagedWidget ("texts" ,boxWidgetClass,infow,NULL,0);
    values= XtCreateManagedWidget ("values",boxWidgetClass,infow,NULL,0);

    XtCreateManagedWidget ("txt1",labelWidgetClass,texts,NULL,0);
    XtCreateManagedWidget ("txt2",labelWidgetClass,texts,NULL,0);
    XtCreateManagedWidget ("txt3",labelWidgetClass,texts,NULL,0);
    XtCreateManagedWidget ("txt4",labelWidgetClass,texts,NULL,0);
    XtCreateManagedWidget ("txt5",labelWidgetClass,texts,NULL,0);
    XtCreateManagedWidget ("txt6",labelWidgetClass,texts,NULL,0);
    XtCreateManagedWidget ("txt7",labelWidgetClass,texts,NULL,0);
    XtCreateManagedWidget ("val1",labelWidgetClass,values,&attribs[n++],1);
    XtCreateManagedWidget ("val2",labelWidgetClass,values,&attribs[n++],1);
    XtCreateManagedWidget ("val3",labelWidgetClass,values,&attribs[n++],1);
    XtCreateManagedWidget ("val4",labelWidgetClass,values,&attribs[n++],1);
    XtCreateManagedWidget ("val5",labelWidgetClass,values,&attribs[n++],1);
    XtCreateManagedWidget ("val6",labelWidgetClass,values,&attribs[n++],1);
    XtCreateManagedWidget ("val7",labelWidgetClass,values,&attribs[n++],1);

    XtAddEventHandler (shell, ButtonPressMask, FALSE, handlerInfoWin, shell);
    XtPopup (shell, XtGrabExclusive);
}

/**************************************************************************/
void ReadInfo (Arg *args) {
    int  n=0;
    char buf[256];
    static char System[64], Version[64];
    static char Cpu[32], Ram[32], Swap[32], Uptime[32];
    char *unknown="unknown";
#ifdef linux
    FILE *fd;

    /****************/
    /* Get hostname */
    if (NULL != (fd = fopen ("/etc/HOSTNAME","rt"))) {
	fgets(buf,256,fd);
	strlrtrim (buf);
	sprintf (Node,"%s",buf);
	fclose (fd);
      } else {
	sprintf (Node,unknown);
	fprintf (stderr,"Cannot open /etc/HOSTNAME file\n");
	}

    /*******************************/
    /* Get system name and version */
    if (NULL != (fd = fopen ("/proc/version","rt"))) {
	char *x, *y;
	fgets(buf,256,fd);
	fclose (fd);
	for (x=buf; *x!=' '; x++) 
	    ;
	*x=0;
	for (x++; *x<'0' || *x>'9'; x++)
	    ;
	y=x;
	for (x++; *x!=' '; x++)
	    ;
	*x=0;
	sprintf (System,"%s",buf);
	sprintf (Version,"%s",y);
      } else {
	sprintf (System,unknown);
	sprintf (Version,unknown);
	fprintf (stderr,"Cannot open /proc/version file\n");
	}

    /****************************/
    /* get CPU type information */
    if (NULL != (fd = fopen ("/proc/cpuinfo","rt"))) {
	while (fgets(buf,256,fd)) {
	    if (1==sscanf (buf,"cpu : %s",Cpu))
		break;
	    }
	fclose (fd);
      } else {
	sprintf (Cpu,unknown);
	fprintf (stderr,"Cannot open /proc/cpuinfo file\n");
	}
    
    /********************************/
    /* get Ram and Swap information */  
    if (NULL != (fd = fopen ("/proc/meminfo","rt"))) {
	while (fgets(buf,256,fd)) {
	    int a,b,c,d,e;
	    if (5==sscanf (buf,"Mem: %i %i %i %i %i",&a,&b,&c,&d,&e)) 
		 sprintf (Ram,"%1.1f MB",(double)a/(double)1048576);
	    if (3==sscanf (buf,"Swap: %i %i %i",&a,&b,&c))
		 sprintf (Swap,"%1.1f MB",(double)a/(double)1048576);
	    }
	fclose (fd);
      } else {
	sprintf (Ram, unknown);
	sprintf (Swap,unknown);
	fprintf (stderr, "Cannot open /proc/meminfo file\n");
	}

    /***********************/
    /* Get time since boot */
    if (NULL != (fd = fopen ("/proc/uptime","rt"))) {
	int d,h,m,s,x;
	int uptime, idletime;
	fgets (buf, 256, fd);
	sscanf (buf,"%i.%i %i.%i",&uptime,&x, &idletime,&x);
	d = uptime/86400;
	uptime -= d*86400;
	h = uptime/3600;
	uptime -= h*3600;
	m = uptime/60;
	uptime -= m*60;
	s = uptime;
	sprintf (Uptime,"%1i days %1i:%02i:%02i",d,h,m,s);
      } else {
	sprintf (Uptime, unknown);
	fprintf (stderr, "Cannot open /proc/uptime file\n");
	}


#endif
#ifdef SVR4 
    /****************/
    {
    struct nlist nl[] = { {"utsname"}, {0} };
    struct utsname un;
    nlist (UNIX_FILE, nl);
    if (nl[0].n_value == 0) {
	sprintf (Node,unknown);
	sprintf (System,unknown);
	sprintf (Version,unknown);
	fprintf (stderr,"Cannot find structure utsname in %s file\n",UNIX_FILE);
      } else {
	lseek   (kmem_fd,nl[0].n_value,SEEK_SET);
	read    (kmem_fd,&un,sizeof(un));
	sprintf (Node,un.nodename);
	sprintf (System,un.sysname);
	sprintf (Version,"%s %s",un.release,un.version);
	}
    }
    
    /****************/
    /* Get CPU type */
    {
    struct nlist nl[] = { {"architecture"}, {0} };
    nlist (UNIX_FILE, nl);
    if (nl[0].n_value == 0) {
	sprintf (Cpu,unknown);
	fprintf (stderr,"Cannot find structure utsname in %s file\n",UNIX_FILE);
      } else {
	lseek   (kmem_fd,nl[0].n_value,SEEK_SET);
	read    (kmem_fd,buf,32);
	sprintf (Cpu,buf);
	}
    }
    
    sprintf (Ram,unknown);
    sprintf (Swap,unknown);
    
    /***********************/
    /* Get time since boot */
    {
    int d,h,m,s;
    int uptime = sa_new.lbolt/100;
    d = uptime/86400;
    uptime -= d*86400;
    h = uptime/3600;
    uptime -= h*3600;
    m = uptime/60;
    uptime -= m*60;
    s = uptime;
    sprintf (Uptime,"%1i days %1i:%02i:%02i",d,h,m,s);
    }
    
#endif
    /**********************/
    /* Set values to args */
    XtSetArg (args[n], XtNlabel, Node); n++;
    XtSetArg (args[n], XtNlabel, System); n++;
    XtSetArg (args[n], XtNlabel, Version); n++;
    XtSetArg (args[n], XtNlabel, Cpu); n++;
    XtSetArg (args[n], XtNlabel, Ram); n++;
    XtSetArg (args[n], XtNlabel, Swap); n++;
    XtSetArg (args[n], XtNlabel, Uptime); n++;
}

/**************************************************************************/
char *strlrtrim (char *b)
{
char  x[128];
char *p=x;
char *r;
strcpy(x,b);
while (*p<=32 && *p!=0)
    p++;
r=p;
while (*r > 32)
    r++;
*r=0;
strcpy(b,p);
return b;
}


