/*
 * This is free software
 *
 * xsar      display basic system load characteristics in window. It can 
 *           watch run queue, cpu load and memory usage.
 * 
 * Author:   Petr Bravenec
 *
 * Version:  0.2 17-Jul-1996
 *
 */
 
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/MenuButton.h>
#include "xsar.h"

static String xsar_res [] = {
   "Xsar*core.height:               160",
   "Xsar*core.width:                540",
   "Xsar*menu.showGrip:             False",
   "Xsar*Quit.label:                Quit",
   "Xsar*Graf.label:                Graph",
   "Xsar*Info.label:                Info",
   "Xsar*Graf.menuName:             menuGraph",
   "Xsar*pozn.borderWidth:          0",
   "Xsar*run_queue.label:           Run Queue",
   "Xsar*cpu_time.label:            CPU load",
   "Xsar*cpu_sys.label:             CPU sys",
   "Xsar*cpu_usr.label:             CPU usr",
   "Xsar*cpu_wait.label:            CPU wait",
   "Xsar*ram_free.label:            Free RAM",
   "Xsar*swap_free.label:           Free swap",
   "Xsar*vm_free.label:             Free virtual memory",
   "Xsar*memory_free.label:         Free memory in MB",
   "Xsar*infow.orientation:         horizontal",
   "Xsar*texts.orientation:         vertical",
   "Xsar*texts*width:               140",
   "Xsar*texts*justify:             Right",
   "Xsar*values.orientation:        vertical",
   "Xsar*values*width:              130",
   "Xsar*values*justify:            Left",
   "Xsar*InfoWin*showGrip:          False",
   "Xsar*InfoWin*txt1.label:        Node name:",
   "Xsar*InfoWin*txt2.label:        System Name:",
   "Xsar*InfoWin*txt3.label:        System version:",
   "Xsar*InfoWin*txt4.label:        CPU type:",
   "Xsar*InfoWin*txt5.label:        RAM size:",
   "Xsar*InfoWin*txt6.label:        Swap size:",
   "Xsar*InfoWin*txt7.label:        Time since boot:",
   "Xsar*InfoWin*borderWidth:       0",
   "Xsar*values*label:              unknown",
   "Xsar*showv.orientation:         horizontal",
   "Xsar*ShowValues*borderWidth:    0",
   "Xsar*ShowValues*txt1.label:     Time",
   "Xsar*ShowValues*txt2.label:     Run queue",
   "Xsar*ShowValues*txt3.label:     %CPU idle",
   "Xsar*ShowValues*txt4.label:     %CPU user",
   "Xsar*ShowValues*txt5.label:     %CPU sys",
   "Xsar*ShowValues*txt6.label:     %CPU wait",
   "Xsar*ShowValues*txt7.label:     Free RAM",
   "Xsar*ShowValues*txt8.label:     Free swap",
   "Xsar*ShowValues*txt9.label:     Free memory",
   NULL
};

#define BORDER 5
#define GOPT_RQUE       0x0001
#define GOPT_CPU        0x0002
#define GOPT_SYS        0x0004
#define GOPT_USR        0x0008
#define GOPT_WAIT       0x0010
#define GOPT_RAM        0x0020
#define GOPT_SWAP       0x0040
#define GOPT_VMFREE     0x0080
#define GOPT_PGIN       0x0100
#define GOPT_PGOUT      0x0200

/********************/
static XtAppContext    kontext;
static Display        *display;
static Window          drawable;
static Widget          xsarwidget;
static Widget          core;
static Widget          comm;
static int             screen;
static GC              GCgraph;
static GC              GCaxis;
static int             coreHeight, coreWidth;
static int             Goptions;
static int             column;
static int             axis_memory=1;
static int             axis_runque=1;
static int             axis_cpu=1;
static int             axis_paging=1;
       int             delay=10;
	
/********************/
static void RepaintWindow   (int);
static void CreateXsarGC    (Widget);
static void CreateInterface (Widget);
static void handlerExpose   (Widget, XtPointer, XEvent *, Boolean *);
static void handlerConfigure(Widget, XtPointer, XEvent *, Boolean *);
static void handlerMouse    (Widget, XtPointer, XEvent *, Boolean *);
static void procTimer       (XtPointer clien_data, XtIntervalId *timer);
static void QuitThisApp     (Widget, XtPointer, XtPointer);
static void Gopt            (Widget, XtPointer, XtPointer);
static void PaintGraph      (int, int);
static void PaintAxis       ();

/**************************************************************************/
int main (int argv, char **argc) {
    startsadc();
    xsarwidget = XtAppInitialize (&kontext, "Xsar",
		 (XrmOptionDescRec *)NULL, ZERO, &argv, argc,
		 xsar_res, (Arg *)NULL, ZERO);
    CreateInterface (xsarwidget);
    XtRealizeWidget (xsarwidget);
    CreateXsarGC    (xsarwidget);
    AddQuitOptionToWindowManager (xsarwidget);
    XtAppAddTimeOut (kontext, 1000*delay, procTimer, NULL);
    XtAppMainLoop   (kontext);
    return 0;
}

/**************************************************************************/
static void CreateInterface (Widget parent) {
    Widget plac, menu, graf, info, quit, ouva;

    /* create main objects in window Xsar */
    plac = XtCreateManagedWidget ("plac",panedWidgetClass,parent,NULL,ZERO);
    menu = XtCreateManagedWidget ("menu",boxWidgetClass,plac,NULL,ZERO);
    core = XtCreateManagedWidget ("core",coreWidgetClass,plac,NULL,ZERO);
    ouva = XtCreateManagedWidget ("ouva",boxWidgetClass,plac,NULL,ZERO);
    comm = XtCreateManagedWidget ("comm",labelWidgetClass,ouva,NULL,ZERO);
    XtAddEventHandler (core, ExposureMask, FALSE, handlerExpose, NULL);
    XtAddEventHandler (core, StructureNotifyMask, FALSE, handlerConfigure, NULL);
    XtAddEventHandler (core, ButtonReleaseMask|ButtonPressMask|LeaveWindowMask, 
						  FALSE, handlerMouse, NULL);
    
    /* create buttons in main menu */ {
    Arg args;
    XtSetArg (args,XtNlabel,(XtArgVal)Node);
    quit = XtCreateManagedWidget ("Quit",commandWidgetClass,menu,NULL,ZERO);
    graf = XtCreateManagedWidget ("Graf",menuButtonWidgetClass,menu,NULL,ZERO);
    info = XtCreateManagedWidget ("Info",commandWidgetClass,menu,NULL,ZERO);
	   XtCreateManagedWidget ("pozn",labelWidgetClass,menu,&args,1);
    XtAddCallback (quit, XtNcallback, QuitThisApp, parent);
    XtAddCallback (info, XtNcallback, ShowInfo, info);
    }

    /* create menu for graph */ {
    Widget mnuG, opt1, opt2, opt3, opt4, opt6, opt7, opt8, opt9, opt10;
#ifndef linux
    Widget opt5;
#endif
    mnuG = XtCreatePopupShell ("menuGraph",simpleMenuWidgetClass,graf,NULL,ZERO);
    opt1 = XtCreateManagedWidget ("run_queue",smeBSBObjectClass, mnuG,NULL,ZERO);
	   XtCreateManagedWidget ("lajna1",   smeLineObjectClass,mnuG,NULL,ZERO);
    opt2 = XtCreateManagedWidget ("cpu_time", smeBSBObjectClass, mnuG,NULL,ZERO);
    opt3 = XtCreateManagedWidget ("cpu_sys",  smeBSBObjectClass, mnuG,NULL,ZERO);
    opt4 = XtCreateManagedWidget ("cpu_usr",  smeBSBObjectClass, mnuG,NULL,ZERO);
    #ifndef linux
    opt5 = XtCreateManagedWidget ("cpu_wait", smeBSBObjectClass, mnuG,NULL,ZERO);
    #endif
	   XtCreateManagedWidget ("lajna1",   smeLineObjectClass,mnuG,NULL,ZERO);
    opt6 = XtCreateManagedWidget ("ram_free", smeBSBObjectClass, mnuG,NULL,ZERO);
    opt7 = XtCreateManagedWidget ("swap_free",smeBSBObjectClass, mnuG,NULL,ZERO);
    opt8 = XtCreateManagedWidget ("vm_free",  smeBSBObjectClass, mnuG,NULL,ZERO);
	   XtCreateManagedWidget ("lajna1",   smeLineObjectClass,mnuG,NULL,ZERO);
    opt9 = XtCreateManagedWidget ("pg_in",    smeBSBObjectClass, mnuG,NULL,ZERO);
    opt10= XtCreateManagedWidget ("pg_out",   smeBSBObjectClass, mnuG,NULL,ZERO);
    XtAddCallback (opt1, XtNcallback, Gopt, (XtPointer)GOPT_RQUE);
    XtAddCallback (opt2, XtNcallback, Gopt, (XtPointer)GOPT_CPU);
    XtAddCallback (opt3, XtNcallback, Gopt, (XtPointer)GOPT_SYS);
    XtAddCallback (opt4, XtNcallback, Gopt, (XtPointer)GOPT_USR);
#ifndef linux
    XtAddCallback (opt5, XtNcallback, Gopt, (XtPointer)GOPT_WAIT);
#endif
    XtAddCallback (opt6, XtNcallback, Gopt, (XtPointer)GOPT_RAM);
    XtAddCallback (opt7, XtNcallback, Gopt, (XtPointer)GOPT_SWAP);
    XtAddCallback (opt8, XtNcallback, Gopt, (XtPointer)GOPT_VMFREE);
    XtAddCallback (opt9, XtNcallback, Gopt, (XtPointer)GOPT_PGIN);
    XtAddCallback (opt10,XtNcallback, Gopt, (XtPointer)GOPT_PGOUT);
    Goptions=1;
    }
}

/**************************************************************************/
static void QuitThisApp (Widget w, XtPointer c, XtPointer x) /* ARGSUSED0 */ {
    XtDestroyApplicationContext (kontext);
    exit(0);
}

/**************************************************************************/
static void Gopt (Widget w, XtPointer c, XtPointer x) /* ARGSUSED0 */ {
    int optmem=GOPT_RAM|GOPT_SWAP|GOPT_VMFREE;
    int optcpu=GOPT_CPU|GOPT_WAIT|GOPT_SYS|GOPT_USR;
    int optque=GOPT_RQUE;
    int optpg =GOPT_PGIN|GOPT_PGOUT;
    if (Goptions & (int)c) {
	Goptions &= ~(int)c; /* switch off this graph */
	if (!Goptions)
	    Goptions = (int)c;
	goto EndGopt;
	}
    if ( (((int)c & optmem) && (Goptions & ~optmem)) ||
	 (((int)c & optcpu) && (Goptions & ~optcpu)) ||
	 (((int)c & optpg ) && (Goptions & ~optpg )) ||
	 (((int)c & optque) && (Goptions & ~optque)) ) {
	Goptions = 0;
	}
    Goptions |= (int)c; /* switch on this graph */
EndGopt:
    XClearWindow (display, drawable);
    RepaintWindow (0);
}

/**************************************************************************/
static void CreateXsarGC    (Widget w) {
    XGCValues gc_attribs;
    long  m1 = GCForeground;
    display  = XtDisplay (w);
    screen   = XScreenNumberOfScreen(XtScreen(w));
    drawable = XtWindow  (core);
    gc_attribs.foreground = WhitePixel (display, screen);
    gc_attribs.line_style = LineOnOffDash;
    GCgraph = XCreateGC (display, drawable, m1, &gc_attribs);
    gc_attribs.foreground = BlackPixel (display, screen);
    GCaxis  = XCreateGC (display, drawable, m1, &gc_attribs);
}

/**************************************************************************/
static void procTimer (XtPointer clien_data, XtIntervalId *timer) /* ARGSUSED */ {
    XtAppAddTimeOut (kontext, 1000*delay, procTimer, NULL);
    readvalues();
    RepaintWindow (9999);
}

/**************************************************************************/
static void handlerMouse (Widget w, XtPointer client_data, XEvent *event, Boolean *continues) /* ARGSUSED */ {
    static int pressed;
    int x;
    switch (event->type) {
	case LeaveNotify:   pressed=0; return;
	case ButtonPress:   pressed=1; return;
	case ButtonRelease: 
	    if (!pressed) 
	       return;
	    pressed=0;
	    break;
	default:
	    fprintf (stderr, "Server X sends me something wrong\n");
	    break;
	}
    x = event->xbutton.x;
    if (x < BORDER)
	x = BORDER;
    if (x > column+BORDER)
	x = column+BORDER;
    x -= BORDER;
    x = sarend - column + x;
    x = (x<0) ? x+SARNUM : x;
/*    
    printf ("x: %1i y: %1i col: %1i sarend: %1i sarnum: %1i vysl: %1i\n",
	    event->xbutton.x, event->xbutton.y,
	    column, sarend, sarnum, x);
*/
    ShowValues (w,event->xbutton.x,event->xbutton.y,&sar[x]);
    return;
}

/**************************************************************************/
static void handlerConfigure (Widget w, XtPointer client_data, XEvent *event, Boolean *continues) /* ARGSUSED */ {
    XWindowAttributes xwa;
    XGetWindowAttributes (display, XtWindow(w), &xwa);
    coreHeight = xwa.height;
    coreWidth  = xwa.width;
    column = coreWidth-BORDER*2;
    if (column > sarnum)
	column = sarnum;
    XClearWindow (display, drawable);
    RepaintWindow (0);
}

/**************************************************************************/
static void RepaintWindow (int count) {
    char u[256]; int l=0;
    XEvent event;
    event.type = Expose;
    event.xexpose.count = count;
    event.xexpose.window = drawable;
    event.xexpose.display = display;
    if (Goptions&GOPT_RAM)    l+=sprintf(u+l,"RAM "); 
    if (Goptions&GOPT_SWAP)   l+=sprintf(u+l,"SWAP ");
    if (Goptions&GOPT_VMFREE) l+=sprintf(u+l,"VMFREE ");
    if (Goptions&GOPT_CPU)    l+=sprintf(u+l,"%%CPU ");
    if (Goptions&GOPT_WAIT)   l+=sprintf(u+l,"%%WAIT ");
    if (Goptions&GOPT_SYS)    l+=sprintf(u+l,"%%SYS ");
    if (Goptions&GOPT_USR)    l+=sprintf(u+l,"%%USR ");
    if (Goptions&GOPT_RQUE)   l+=sprintf(u+l,"RUN QUEUE ");
    if (Goptions&GOPT_PGIN)   l+=sprintf(u+l,"PgIn ");
    if (Goptions&GOPT_PGOUT)  l+=sprintf(u+l,"PgOut ");
    if (Goptions&(GOPT_RAM|GOPT_SWAP|GOPT_VMFREE)) {
        int r=axis_memory;
        sprintf (u+l,"line = %1i MB",(r<10)?1:(r<100)?10:(r<1000)?100:1000); 
        }
    if (Goptions&(GOPT_PGIN|GOPT_PGOUT)) {
        int r=axis_paging;
        sprintf (u+l,"line = %1i pgs",(r<10)?1:(r<100)?10:(r<1000)?100:1000); 
        }
    if (Goptions&GOPT_RQUE) {
        int r=axis_runque;
        sprintf (u+l,"line = %1i",(r<10)?1:(r<100)?10:(r<1000)?100:1000); 
        }
    XtVaSetValues (comm, XtNlabel, u, NULL);
    XSendEvent (display, drawable, TRUE, ExposureMask, &event);
}

/**************************************************************************/
static void handlerExpose (Widget w, XtPointer client_data, XEvent *event, Boolean *continues) /* ARGSUSED0 */ {
    if (event->type==Expose && event->xexpose.count==9999) {
	if (sarnum < 2)
	    return;
	if (++column+BORDER > coreWidth-BORDER) {
	    column -= 50;
	    XClearWindow  (display, drawable);
	    RepaintWindow (0);
	    return;
	    }
	PaintGraph (sarend, column);
	}
    if (event->type==Expose && event->xexpose.count==0) {
	int  i,j;
	PaintAxis ();
	if (sarnum < 2)
	    return;
	j= (sarend>=column) ? sarend-column : sarend-column+SARNUM;
	if (j==0)
	    j=1;
	for (i=1; i<=column; i++,j++) {
	    if (j>=SARNUM)
		j=0;
	    PaintGraph (j,i);
	    }
	}
}

/**************************************************************************/
static void PaintAxis () {
    int cH=coreHeight- BORDER*2;
    int cW=coreWidth - BORDER*2;
    int i,rng=1;
    if (Goptions & GOPT_RQUE)
	rng = axis_runque;
    if (Goptions & (GOPT_CPU|GOPT_SYS|GOPT_USR|GOPT_WAIT))
	rng = axis_cpu;
    if (Goptions & (GOPT_RAM|GOPT_SWAP|GOPT_VMFREE)) 
	rng = axis_memory;
    if (Goptions & (GOPT_PGIN|GOPT_PGOUT)) 
	rng = axis_paging;
    for (i=0; i<=rng; i+=(rng<10)?1:(rng<100)?10:(rng<1000)?100:1000) {
	int x;
	x = cH-(i*cH/rng);
/*        
	printf ("i=%1i, rng=%1i, x=%1i, cH=%1i\n",i,rng,x,cH);
*/        
	XDrawLine (display, drawable, GCaxis, 
		     BORDER, x+BORDER, 
		  cW+BORDER, x+BORDER);
	}
}

/**************************************************************************/
static void PaintGraph (int idx, int col) {
    int  fy,ty;
    int  ido = (0==idx) ? SARNUM-1: idx-1;
    int  cH = coreHeight - BORDER*2;
    Sar *so=&sar[ido];
    Sar *sn=&sar[idx];
    /* Draw time stamp to graph */
    if (sar[idx].time%3600 == 0) { /* hole hour */
	XDrawLine (display, drawable, GCaxis, col+BORDER, 0, 
					      col+BORDER, coreHeight);
	}
    if (sar[idx].time%900 == 0) { /* quorter of hour */
	XDrawLine (display, drawable, GCaxis, col+BORDER, coreHeight-BORDER, 
					       col+BORDER, coreHeight);
	}
    /********************************************/
    if (Goptions & GOPT_RQUE) {
	if (sn->runque > axis_runque*100) {
	    axis_runque = (sn->runque+100)/100;
	    goto ClearAndRepaint;
	    }
	fy = (cH - (cH * so->runque)/100 /axis_runque ) + BORDER;
	ty = (cH - (cH * sn->runque)/100 /axis_runque ) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    /********************************************/
    if (Goptions & GOPT_CPU) {
	fy = (cH - cH * (10000-so->cpu_idle)/10000) + BORDER;
	ty = (cH - cH * (10000-sn->cpu_idle)/10000) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    if (Goptions & GOPT_SYS) {
	fy = (cH - cH * (so->cpu_sys)/10000) + BORDER;
	ty = (cH - cH * (sn->cpu_sys)/10000) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    if (Goptions & GOPT_USR) {
	fy = (cH - cH * (so->cpu_user)/10000) + BORDER;
	ty = (cH - cH * (sn->cpu_user)/10000) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    if (Goptions & GOPT_WAIT) {
	fy = (cH - cH * (so->cpu_wait)/10000) + BORDER;
	ty = (cH - cH * (sn->cpu_wait)/10000) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    /********************************************/
    if (Goptions & GOPT_RAM) {
	if (sn->freemem > axis_memory*100) {
	    axis_memory = (sn->freemem<1000) ? 
			  (sn->freemem+100)/100   : 
                        (((sn->freemem+100)/100)+10)/10*10;
	    goto ClearAndRepaint;
	    }
	fy = (cH - cH * so->freemem/axis_memory/100 ) + BORDER;
	ty = (cH - cH * sn->freemem/axis_memory/100 ) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    /********************************************/
    if (Goptions & GOPT_SWAP) {
	if (sn->freeswap > axis_memory*100) {
	    axis_memory = (sn->freeswap<1000) ? 
			  (sn->freeswap+100)/100   : 
                        (((sn->freeswap+100)/100)+10)/10*10;
	    goto ClearAndRepaint;
	    }
	fy = (cH - cH * so->freeswap/axis_memory/100 ) + BORDER;
	ty = (cH - cH * sn->freeswap/axis_memory/100 ) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    /********************************************/
    if (Goptions & GOPT_VMFREE) {
	if (sn->free > axis_memory*100) {
	    axis_memory = (sn->free<1000) ? 
			  (sn->free+100)/100   : 
                        (((sn->free+100)/100)+10)/10*10;
	    goto ClearAndRepaint;
	    }
	fy = (cH - cH * so->free/axis_memory/100 ) + BORDER;
	ty = (cH - cH * sn->free/axis_memory/100 ) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    /********************************************/
    if (Goptions & GOPT_PGIN) {
	if (sn->pg_in > axis_paging) {
	    axis_paging = (sn->pg_in<10) ? 
			   sn->pg_in+1   : 
                          (sn->pg_in<100) ?
			  (sn->pg_in+10)/10*10 :
                          (sn->pg_in<1000) ?
			  (sn->pg_in+100)/100*100 :
			  (sn->pg_in+1000)/1000*1000 ;
	    goto ClearAndRepaint;
	    }
	fy = (cH - cH * so->pg_in/axis_paging ) + BORDER;
	ty = (cH - cH * sn->pg_in/axis_paging ) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}
    if (Goptions & GOPT_PGOUT) {
	if (sn->pg_out > axis_paging) {
	    axis_paging = (sn->pg_out<10) ? 
			   sn->pg_out+1   : 
                          (sn->pg_out<100) ?
			  (sn->pg_out+10)/10*10 :
                          (sn->pg_out<1000) ?
			  (sn->pg_out+100)/100*100 :
			  (sn->pg_out+1000)/1000*1000 ;
	    goto ClearAndRepaint;
	    }
	fy = (cH - cH * so->pg_out/axis_paging ) + BORDER;
	ty = (cH - cH * sn->pg_out/axis_paging ) + BORDER;
	XDrawLine (display, drawable, GCgraph, col-1+BORDER, fy, col+BORDER, ty);
	}

    return;
    
ClearAndRepaint:
    if (axis_memory > 10)  axis_memory = axis_memory/10*10;
    if (axis_paging >1000) axis_paging = axis_paging/1000*1000;
    if (axis_paging >100 ) axis_paging = axis_paging/100 *100 ;
    if (axis_paging >10  ) axis_paging = axis_paging/10  *10  ;
/*        
    printf ("Axis runque: %1i cpu: %1i memory: %1i idx: %1i\n",
	  axis_runque,
	  axis_cpu,
	  axis_memory,
	  idx
	  );
*/
    XClearWindow  (display, drawable);
    RepaintWindow (0);
    return;
}

